/* ------------------------------------------------------------------------- */
/*   rivatv-driver.c Video4Linux driver for NVIDIA display adapters	     */
/* ------------------------------------------------------------------------- */
/*   Copyright (C) 2000 Ferenc Bakonyi <fero@drama.obuda.kando.hu>
 *   Based on various sources by Alan Cox
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *   
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *   
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.		     */
/* ------------------------------------------------------------------------- */

/* Module versioning */
#include <linux/config.h>
#if defined (CONFIG_MODVERSIONS) && !defined (MODVERSIONS)
#define MODVERSIONS
#endif
#include <linux/version.h>
#ifdef MODVERSIONS
#include <linux/modversions.h>
#endif
/* End module versioning */

#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/init.h>
#include <linux/poll.h>
#include <linux/pci.h>
#include <linux/video_decoder.h>
#include <linux/videotext.h>

#include "i2c-riva.h"
#include "rivatv.h"
#include "tuner.h"
#include "audiochip.h"

/*----------------------------------------------------------------------
 *
 * Video4Linux video capture device driver part
 *
 *----------------------------------------------------------------------*/

extern int  init_i2c_riva (void);
extern int  cleanup_i2c_riva (void);
extern int  rivatv_video_start (struct rivatv_info *);
extern int  rivatv_video_stop (struct rivatv_info *);
extern void rivatv_overlay_start (struct rivatv_info *);
extern void rivatv_overlay_stop (struct rivatv_info *);
extern void rivatv_reset_queue (struct rivatv_info *);
extern int  rivatv_add_queue (struct rivatv_info *, struct video_mmap *);
extern int  rivatv_sync (struct rivatv_info *, int);
extern int  rivatv_check_format (struct rivatv_info *, struct video_mmap *);
extern int  rivatv_check_source (struct rivatv_info *, struct video_mmap *);
extern void rivatv_get_capability (struct rivatv_info *, struct video_capability *);
extern void rivatv_set_overlay_format (struct rivatv_info *);
extern int  rivatv_mmap (struct rivatv_info *, struct vm_area_struct *, ulong, ulong);

/* Init */

/* Setup video picture once */

static void __init rivatv_setup_picture (struct rivatv_info *info)
{
	info->picture.brightness = 32768;
	info->picture.hue	 = 32768;
	info->picture.colour	 = 32768;
	info->picture.contrast	 = 32768;
	info->picture.whiteness	 = 32768;
	info->picture.depth	 = 24;
	info->picture.palette	 = VIDEO_PALETTE_RGB24;
}

/* Definition of available video sources. */

#define RIVATV_SOURCES	    2
#define RIVATV_TUNER_SOURCE 2

static struct video_channel rivatv_channel[RIVATV_SOURCES + 1] = {
	{ 0, "Composite",  0, 0,	      VIDEO_TYPE_CAMERA, VIDEO_MODE_AUTO },
	{ 1, "S-Video",	   0, 0,	      VIDEO_TYPE_CAMERA, VIDEO_MODE_AUTO },
	{ 2, "Television", 1, VIDEO_VC_TUNER, VIDEO_TYPE_TV,	 VIDEO_MODE_AUTO }
};

static struct rivatv_source rivatv_source[] = {
	{ I2C_DRIVERID_SAA7111A, 
	  { 3, 4, -1 }, 
	  { VIDEO_PALETTE_UYVY, VIDEO_PALETTE_RGB24, VIDEO_PALETTE_RGB565, 
	    VIDEO_PALETTE_YUV422, VIDEO_PALETTE_YUYV, VIDEO_PALETTE_YUV411, -1 } ,
	  VIDEO_PALETTE_UYVY
	},
	{ I2C_DRIVERID_SAA7113,
	  { 3, 6, 1 },
	  { VIDEO_PALETTE_UYVY, -1 } ,
	  VIDEO_PALETTE_UYVY
	},
	{ I2C_DRIVERID_SAA7108,
	  { 0, 6, 0 },
	  { VIDEO_PALETTE_UYVY, -1 } ,
	  VIDEO_PALETTE_UYVY
	},
	{ I2C_DRIVERID_VPX32XX,
	  { 1, 7, -1 },
	  { VIDEO_PALETTE_UYVY, -1 } ,
	  VIDEO_PALETTE_UYVY
	},
	{ I2C_DRIVERID_TW98,
	  { 2, 5, -1 },
	  { VIDEO_PALETTE_UYVY, -1 } ,
	  VIDEO_PALETTE_UYVY
	},
	{ -1 }
};

/* Setup the video decoder source for a given I2C chip id. */
static int rivatv_setup_source (struct rivatv_info *info, int i2c_id)
{
	struct rivatv_source *source = rivatv_source;

	/* determine number of input source */
	info->sources = RIVATV_SOURCES;
	if (info->tuner_required != -1 && info->i2c->tuner)
		info->sources++;

	/* go through list of possible decoder chips */
	while (source->id != -1) {
		if (source->id == i2c_id) {
			info->source = source;
			/* copy default values for input modes if card has not 
			   been detected correctly */
			if (info->card.identified == 0)
				memcpy (info->card.mode, source->mode, sizeof (info->card.mode));
			return 0;
		}
		source++;
	}
	return -1;
}

static int __init rivatv_video_init (void)
{
	int err;

	rivatv_setup_picture (riva_board);
	err = init_i2c_riva ();
	if (err)
		PRINTK_ERR ("error initializing I2C access\n");
	return err;
}

/* Standard character-device-driver functions */

#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 5, 0)
static int rivatv_video_open (struct video_device *dev, int nb)
{
#else
static int rivatv_video_open  (struct inode *inode, struct file *file)
{
        struct video_device *dev = video_devdata (file);
#endif
	struct rivatv_info *info = (struct rivatv_info *) dev->priv;

	/* open video device just once */
	if (info->device_busy) {
		DPRINTK ("video device busy: %d\n", info->device_busy);
		return -EBUSY;
	}

	/* check for the video decoder device */
	if (info->i2c == NULL || info->i2c->video_decoder == NULL) {
		DPRINTK ("no video decoder device registered\n");
		return -ENODEV;
	}

	info->device_busy++;
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 5, 0)
	file->private_data = dev;
#endif

	/* setup decoder source */
	if (info->source == NULL) {
		if (rivatv_setup_source (info, info->i2c->video_decoder->driver->id)) {
			PRINTK_ERR ("no source defined for decoder chip 0x%04x\n", 
				    info->i2c->video_decoder->driver->id);
		}
	}

	MOD_INC_USE_COUNT;

	/* increment module usage counter of I2C clients */
	i2c_inc_use_client (info->i2c->video_decoder);
	if (info->tuner_required != -1 && info->i2c->tuner != NULL) {
		i2c_inc_use_client (info->i2c->tuner);
		info->tuner_used++;
		info->tuner_selected = 0;
	}
	if (info->audio_required != -1) {
		if (info->i2c->audio_decoder != NULL) {
			i2c_inc_use_client (info->i2c->audio_decoder);
			info->audio_decoder_used++;
		}
		if (info->i2c->audio_processor != NULL) {
			i2c_inc_use_client (info->i2c->audio_processor);
			info->audio_processor_used++;
		}
	}
	return 0;
}

#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 5, 0)
static void rivatv_video_close (struct video_device *dev)
{
#else
static int rivatv_video_close  (struct inode *inode, struct file *file)
{
        struct video_device *dev = video_devdata (file);
#endif
	struct rivatv_info *info = (struct rivatv_info *) dev->priv;

	/* reset the busy counter (see open()) */
	if (info->device_busy > 0)
		info->device_busy--;

	rivatv_overlay_stop (info);
	rivatv_video_stop (info);
	rivatv_reset_queue (info);

	/* reset decoder source */
	info->source = NULL;

	/* decrement module usage counter of I2C clients */
	if (info->tuner_used) {
		i2c_dec_use_client (info->i2c->tuner);
		info->tuner_used--;
		info->tuner_selected = 0;
	}
	if (info->audio_decoder_used) {
		i2c_dec_use_client (info->i2c->audio_decoder);
		info->audio_decoder_used--;
	}
	if (info->audio_processor_used) {
		int input = AUDIO_MUTE;
		info->i2c->audio_processor->driver->command (info->i2c->audio_processor, AUDC_SET_INPUT, &input);
		i2c_dec_use_client (info->i2c->audio_processor);
		info->audio_processor_used--;
	}
	i2c_dec_use_client (info->i2c->video_decoder);

	MOD_DEC_USE_COUNT;
#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 5, 0)
	return;
#else
	file->private_data = NULL;
	return 0;
#endif
}


#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 5, 0)
static int rivatv_video_ioctl (struct video_device *dev, unsigned int cmd, void *arg)
{
#else
static int rivatv_video_ioctl (struct inode *inode, struct file *file, unsigned int cmd, unsigned long a)
{
        struct video_device *dev = video_devdata (file);
	void *arg = (void *) a;
#endif
	struct rivatv_info *info = (struct rivatv_info *) dev->priv;
	struct i2c_client *decoder = info->i2c->video_decoder;
	struct i2c_client *tuner = info->i2c->tuner;
	struct i2c_client *audio_decoder = info->i2c->audio_decoder;
	struct i2c_client *audio_processor = info->i2c->audio_processor;

	switch (cmd) {

	/* Get capabilities */
	case VIDIOCGCAP:
	{
		struct video_capability v;

		DPRINTK ("VIDIOCGCAP\n");
		strcpy (v.name, "rivatv video");
		v.type = VID_TYPE_CAPTURE;
		if (info->chip.arch == NV_ARCH_10 || info->chip.arch == NV_ARCH_20 || info->chip.arch == NV_ARCH_04)
			v.type |= VID_TYPE_OVERLAY | VID_TYPE_CHROMAKEY;
		v.channels  = info->sources;
		v.audios    = 0;
		v.minwidth  = RIVATV_MIN_WIDTH;
		v.minheight = RIVATV_MIN_HEIGHT;

		rivatv_get_capability (info, &v);

		if (copy_to_user (arg, &v, sizeof (v)))
			return -EFAULT;
		break;
	}

	/* Get channel info (sources) */
	case VIDIOCGCHAN:
	{
		struct video_channel v;

		DPRINTK ("VIDIOCGCHAN\n");
		if (copy_from_user (&v, arg, sizeof (v)))
			return -EFAULT;
		if (v.channel >= info->sources)
			return -EINVAL;

		rivatv_channel[v.channel].flags &= ~VIDEO_VC_AUDIO;
		if (v.channel == RIVATV_TUNER_SOURCE) {
			strcpy (rivatv_channel[v.channel].name, tvbox ? "TV-Box" : "Television");
			if (info->audio_required != -1 && audio_processor != NULL)
				rivatv_channel[v.channel].flags |= VIDEO_VC_AUDIO;
		}

		if (copy_to_user (arg, &rivatv_channel[v.channel], sizeof (struct video_channel)))
			return -EFAULT;
		break;
	}

	/* Set channel */
	case VIDIOCSCHAN:
	{
		struct video_channel v;
		int val;

		DPRINTK ("VIDIOCSCHAN\n");
		if (copy_from_user (&v, arg, sizeof (v)))
			return -EFAULT;
		if (v.channel >= info->sources)
			return -EINVAL;

// --> Force channel here

		/* select type of tuner right here if necessary */
		if (v.channel == RIVATV_TUNER_SOURCE) {
			if (!info->tuner_selected) {
				int type = info->tuner_required;
				tuner->driver->command (tuner, TUNER_SET_TYPE, &type);
				info->tuner_selected++;
			}
			tuner->driver->command (tuner, cmd, &v);
		}

		/* set input source of audio chip */
		if (info->audio_required != -1) {
			int input = (v.channel == RIVATV_TUNER_SOURCE) ? AUDIO_TUNER : AUDIO_EXTERN;
			/* the tda8425 processor has 2 input channels, current driver uses 1 only */
			if (audio_processor != NULL)
				audio_processor->driver->command (audio_processor, AUDC_SET_INPUT, &input);
		}

		/* save selected video channel */
		info->channel = v.channel;

		down (&info->decoder_lock);

		/* set input mode of video decoder */
		val = info->card.mode[v.channel];
		decoder->driver->command (decoder, DECODER_SET_INPUT, &val);

		/* set video norm of video decoder */
		val = v.norm;
		decoder->driver->command (decoder, DECODER_SET_NORM, &val);

		/* set output format of video decoder */
		val = info->source->format;
		decoder->driver->command (decoder, DECODER_ENABLE_OUTPUT, &val);

		/* obtain current video standard */
		val = 0;
		decoder->driver->command (decoder, DECODER_GET_STATUS, &val);
		if (val & DECODER_STATUS_PAL)
			info->format.norm = VIDEO_MODE_PAL;
		else if (val & DECODER_STATUS_NTSC)
			info->format.norm = VIDEO_MODE_NTSC;
		else if (val & DECODER_STATUS_SECAM)
			info->format.norm = VIDEO_MODE_SECAM;

		up (&info->decoder_lock);
		break;
	}

	/* Get tuner abilities */
	case VIDIOCGTUNER:
	{
		struct video_tuner v;
		int val = 0;

		DPRINTK ("VIDIOCGTUNER\n");
		if (copy_from_user (&v, arg, sizeof (v)))
			return -EFAULT;
		strcpy (v.name, rivatv_channel[RIVATV_TUNER_SOURCE].name);
		v.rangelow = 44 * 16;
		v.rangehigh = 958 * 16;
		v.flags = VIDEO_TUNER_PAL | VIDEO_TUNER_NTSC | VIDEO_TUNER_SECAM;
		v.mode = info->format.norm;

		down (&info->decoder_lock);
		decoder->driver->command (decoder, DECODER_GET_STATUS, &val);
		up (&info->decoder_lock);
		v.signal = (val & DECODER_STATUS_GOOD) ? 0xffff : 0;

		if (copy_to_user (arg, &v, sizeof (v)))
			return -EFAULT;
		break;
	}

	/* Tune the tuner for the current channel */
	case VIDIOCSTUNER:
	{
		struct video_tuner v;

		DPRINTK ("VIDIOCSTUNER\n");
		if (copy_from_user (&v, arg, sizeof (v)))
			return -EFAULT;

		if (v.tuner != 0)
			return -EINVAL;
				
		if (v.mode != VIDEO_MODE_PAL   && v.mode != VIDEO_MODE_NTSC && 
		    v.mode != VIDEO_MODE_SECAM && v.mode != VIDEO_MODE_AUTO)
			return -EOPNOTSUPP;

		if (info->format.norm != v.mode) {
			int val = v.mode;
			down (&info->decoder_lock);
			decoder->driver->command (decoder, DECODER_SET_NORM, &val);
			val = 0;
			decoder->driver->command (decoder, DECODER_GET_STATUS, &val);
			if (val & DECODER_STATUS_PAL)
				info->format.norm = VIDEO_MODE_PAL;
			else if (val & DECODER_STATUS_NTSC)
				info->format.norm = VIDEO_MODE_NTSC;
			else if (val & DECODER_STATUS_SECAM)
				info->format.norm = VIDEO_MODE_SECAM;
			up (&info->decoder_lock);
		}
		break;
	}

	/* Get picture properties */
	case VIDIOCGPICT:
	{
		struct video_picture *v = &info->picture;

		DPRINTK ("VIDIOCGPICT\n");
		if (copy_to_user (arg, v, sizeof (struct video_picture)))
			return -EFAULT;
		break;
	}

	/* Set picture properties */
	case VIDIOCSPICT:
	{
		struct video_picture v;

		DPRINTK ("VIDIOCSPICT\n");
		if (copy_from_user (&v, arg, sizeof (v)))
			return -EFAULT;

		info->picture = v;
		down (&info->decoder_lock);
		decoder->driver->command (decoder, DECODER_SET_PICTURE, &v);
		up (&info->decoder_lock);
		break;
	}

	/* Start, end capture */
	case VIDIOCCAPTURE:
	{
		int enable;

		if (copy_from_user (&enable, arg, sizeof (enable)))
			return -EFAULT;
		
		DPRINTK ("VIDIOCCAPTURE: %d\n", enable);

		if (enable) {
			rivatv_video_stop (info);
			rivatv_set_overlay_format (info);
			rivatv_video_start (info);
			rivatv_overlay_start (info);
			info->overlay.enabled = 1;
		}
		else {
			rivatv_overlay_stop (info);
			rivatv_video_stop (info);
			info->overlay.enabled = 0;
		}
		break;
	}

	/* Get the video overlay window */
	case VIDIOCGWIN:
	{
		struct video_window *v = &info->overlay.window;

		DPRINTK ("VIDIOCGWIN\n");
		if (copy_to_user (arg, v, sizeof (struct video_window)))
			return -EFAULT;
		break;
	}

	/* Set the video overlay window - passes clip list for hardware smarts, chromakey etc. */
	case VIDIOCSWIN:
	{
		struct video_window window;
		struct video_window *win = &info->overlay.window;
		int restart = 0;

		if (copy_from_user (&window, arg, sizeof (window)))
			return -EFAULT;

		DPRINTK ("VIDIOCSWIN: %s: %dx%d at (%d,%d) (key: %08X)\n", "Set Window",
			 window.width, window.height, window.x, window.y, window.chromakey);

		/* check for a change in overlay arguments */
		if ((window.width != win->width || window.height != win->height ||
		     window.x	  != win->x	|| window.y	 != win->y ||
		     window.chromakey != win->chromakey) && info->overlay.enable ) {
			restart = 1;
		}

		/* save given overlay window arguments */
		memcpy (win, &window, sizeof (window));
		win->flags = 0;
		win->clipcount = 0;
		win->clips = NULL;

		if (restart && (info->chip.arch == NV_ARCH_04 || 
				info->chip.arch == NV_ARCH_10 || 
				info->chip.arch == NV_ARCH_20)) {
			rivatv_overlay_stop (info);
			rivatv_video_stop (info);
			rivatv_set_overlay_format (info);
			rivatv_video_start (info);
			rivatv_overlay_start (info);
		}
		break;
	}

	/* Get frame buffer */
	case VIDIOCGFBUF:
	{
		struct video_buffer *buffer = &info->overlay.buffer;

		buffer->base = (void *) info->base1_region;

		DPRINTK ("VIDIOCGFBUF: %dx%d (%d bits, %d bpl) @ 0x%08x\n",
			 buffer->width, buffer->height, buffer->depth, 
			 buffer->bytesperline, (u32) buffer->base);

		if (copy_to_user (arg, buffer, sizeof (struct video_buffer)))
			return -EFAULT;
		break;
	}

	/* Set frame buffer - root only */
	case VIDIOCSFBUF:
	{
		struct video_buffer *buffer = &info->overlay.buffer;

		if (!capable (CAP_SYS_ADMIN))
			return -EPERM;

		if (copy_from_user (buffer, arg, sizeof (struct video_buffer)))
			return -EFAULT;

		DPRINTK ("VIDIOCSFBUF: %dx%d (%d bits, %d bpl) @ 0x%08x\n",
			 buffer->width, buffer->height, buffer->depth, 
			 buffer->bytesperline, (u32) buffer->base);
		break;
	}
	
	/* Video key event, cuts capture on all DMA windows with this key (0xFFFFFFFF == all) */
	case VIDIOCKEY:
	{
		DPRINTK ("VIDIOCKEY\n");
		return -ENOIOCTLCMD;
	}

	/* Get tuner */
	case VIDIOCGFREQ:
	{
		unsigned long v = info->tuner_frequency;

		DPRINTK ("VIDIOCGFREQ\n");
		if (copy_to_user (arg, &v, sizeof (v)))
			return -EFAULT;
		break;
	}

	/* Set tuner */
	case VIDIOCSFREQ:
	{
		unsigned long v;

		DPRINTK ("VIDIOCSFREQ\n");
		if (copy_from_user (&v, arg, sizeof (v)))
			return -EFAULT;
		info->tuner_frequency = v;

		if (info->tuner_required == -1 || tuner == NULL)
			return -ENODEV;
		if (info->tuner_frequency != 0)
			tuner->driver->command (tuner, cmd, &v);
		break;
	}

	/* Get audio info */
	case VIDIOCGAUDIO:
	{
		struct video_audio v;

		DPRINTK ("VIDIOCGAUDIO\n");
		v = info->audio;
		v.flags &= ~(VIDEO_AUDIO_MUTE | VIDEO_AUDIO_MUTABLE);
		v.flags |= VIDEO_AUDIO_MUTABLE;
		strcpy (v.name, "TV");
		v.mode = VIDEO_SOUND_MONO;

		if (info->audio_required == -1 || audio_processor == NULL)
			return -ENODEV;
		audio_processor->driver->command (audio_processor, cmd, &v);
		if (audio_decoder != NULL)
			audio_decoder->driver->command (audio_decoder, cmd, &v);

		if (copy_to_user (arg, &v, sizeof (v)))
			return -EFAULT;
		break;
	}

	/* Audio source, mute etc */
	case VIDIOCSAUDIO:
	{
		struct video_audio v;
		int input;

		DPRINTK ("VIDIOCSAUDIO\n");
		if (copy_from_user (&v, arg, sizeof (v)))
			return -EFAULT;

		if (info->audio_required == -1 || audio_processor == NULL)
			return -ENODEV;

		input = (info->channel == RIVATV_TUNER_SOURCE) ? AUDIO_TUNER : AUDIO_EXTERN;
		if (v.flags & VIDEO_AUDIO_MUTE)
			input |= AUDIO_MUTE;
		audio_processor->driver->command (audio_processor, AUDC_SET_INPUT, &input);
		audio_processor->driver->command (audio_processor, cmd, &v);
		if (audio_decoder != NULL)
			audio_decoder->driver->command (audio_decoder, cmd, &v);

		info->audio = v;
		break;
	}

	/* Sync with mmap grabbing */
	case VIDIOCSYNC:
	{
		int frame, ret;

		if (copy_from_user (&frame, arg, sizeof (frame)))
			return -EFAULT;
		DPRINTK2 ("VIDIOCSYNC: %d\n", frame);
		ret = rivatv_sync (info, frame);
		return ret;
	}

	/* Grab frames */
	case VIDIOCMCAPTURE:
	{
		struct video_mmap v;
		int ret, restart, format;

		if (copy_from_user (&v, arg, sizeof (v)))
			return -EFAULT;

		DPRINTK2 ("VIDIOCMCAPTURE: %d\n", v.frame);
		format = rivatv_check_source (info, &v);
		if (format) {
			down (&info->decoder_lock);
			decoder->driver->command (decoder, DECODER_ENABLE_OUTPUT, &format);
			up (&info->decoder_lock);
		}

		restart = rivatv_check_format (info, &v);
		if (restart < 0)
			return -EINVAL;
		
		ret = rivatv_add_queue (info, &v);
		if (ret)
			return ret;

		/* start video here */
		if (restart)
			rivatv_video_stop (info);
		rivatv_video_start (info);
		break;
	}

	/* Memory map buffer info */
	case VIDIOCGMBUF:
	{
		struct video_mbuf v;
		int i;

		DPRINTK ("VIDIOCGMBUF\n");
		memset (&v, 0 , sizeof (v));
		v.size = RIVATV_CAPTURE_BUFSIZE * capbuffers;
		v.frames = capbuffers;
		for (i = 0; i < capbuffers; i++)
			v.offsets[i] = i * RIVATV_CAPTURE_BUFSIZE;
		if (copy_to_user (arg, &v, sizeof (v)))
			return -EFAULT;
		break;
	}

	case VIDIOCGUNIT:	/* Get attached units */
	case VIDIOCGCAPTURE:	/* Get subcapture */
	case VIDIOCSCAPTURE:	/* Set subcapture */
	case VIDIOCSPLAYMODE:	/* Set output video mode/feature */
	case VIDIOCSWRITEMODE:	/* Set write mode */
	case VIDIOCGPLAYINFO:	/* Get current playback info from hardware */
	case VIDIOCSMICROCODE:	/* Load microcode into hardware */
	case VIDIOCGVBIFMT:	/* Get VBI information */
	case VIDIOCSVBIFMT:	/* Set VBI information */
	default:
		DPRINTK ("V4L: Requested IOCTL (0x%08X) not implemented\n", cmd);
		return -ENOIOCTLCMD;
	}
	return 0;
}

#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 5, 0)
static int rivatv_video_mmap (struct video_device *dev, const char *adr, unsigned long size)
{
#else
static int rivatv_video_mmap (struct file *file, struct vm_area_struct *vma)
{
        struct video_device *dev = video_devdata (file);
#endif
	struct rivatv_info *info = (struct rivatv_info *) dev->priv;
	int ret;

	spin_lock_irq (&info->video_lock);
#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 5, 0)
	ret = rivatv_mmap (info, NULL, (ulong) adr, size);
#else
	ret = rivatv_mmap (info, vma, vma->vm_start, vma->vm_end - vma->vm_start);
#endif
	spin_unlock_irq (&info->video_lock);
	return ret;
}

/*----------------------------------------------------------------------
 *
 * Video4Linux VBI device driver part
 *
 *----------------------------------------------------------------------*/

/* Standard character-device-driver functions */

#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 5, 0)
static int rivatv_vbi_open (struct video_device *dev, int nb)
{
#else
static int rivatv_vbi_open  (struct inode *inode, struct file *file)
{
        struct video_device *dev = video_devdata (file);
#endif
	struct rivatv_info *info = (struct rivatv_info *) dev->priv;

	if (info->i2c == NULL || info->i2c->video_decoder == NULL)
		return -ENODEV;

	MOD_INC_USE_COUNT;
	i2c_inc_use_client (info->i2c->video_decoder);
	return 0;
}

#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 5, 0)
static void rivatv_vbi_close (struct video_device *dev)
{
#else
static int rivatv_vbi_close  (struct inode *inode, struct file *file)
{
        struct video_device *dev = video_devdata (file);
#endif
	struct rivatv_info *info = (struct rivatv_info *) dev->priv;

	i2c_dec_use_client (info->i2c->video_decoder);
	MOD_DEC_USE_COUNT;
#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 5, 0)
	return;
#else
	return 0;
#endif
}


#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 5, 0)
static int rivatv_vbi_ioctl (struct video_device *dev, unsigned int cmd, void *arg) 
{
#else
static int rivatv_vbi_ioctl (struct inode *inode, struct file *file, unsigned int cmd, unsigned long a)
{
	void *arg = (void *) a;
#endif

	switch (cmd) {
	case VIDIOCGCAP:
	{
		struct video_capability v;
		v.type = VID_TYPE_TELETEXT;
		v.channels = 2;
		v.audios = 0;
		v.maxwidth = v.minwidth = 0;
		v.maxheight = v.minheight = 0;
		strcpy(v.name, "rivatv vbi");
		if (copy_to_user (arg, &v, sizeof(v)))
			return -EFAULT;
		break;
	}
	case VIDIOCGVBIFMT:
	{
		struct vbi_format f;
		f.sampling_rate = 14750000UL; /* these are foobars */
		f.sample_format = VIDEO_PALETTE_RAW;
		f.flags = VBI_INTERLACED;
		if (copy_to_user(arg, &f, sizeof(f)))
			return -EFAULT;
		break;
	}
	case VIDIOCSVBIFMT:
	{
		struct vbi_format f;
		if (copy_from_user(&f, arg,sizeof(f)))
			return -EFAULT;
		DPRINTK( "VIDIOCSVBIINFO(%d,%d,%d,%d,%d,%d,%d,%x)\n", 
			 f.sampling_rate, f.samples_per_line, f.sample_format,
			 f.start[0], f.start[1], f.count[0], f.count[1],
			 f.flags);
		break;
	}
	default:
		DPRINTK( "VBI: requested IOCTL (%d) not implemented\n",cmd);
		return -ENOIOCTLCMD;
	}
	return 0;
}

/* Externel references. */

extern struct pci_device_id rivatv_pci_table[];
extern struct pci_driver rivatv_driver;
extern int  rivatv_register_v4l (const struct rivatv_initdata *);
extern void rivatv_unregister_v4l (void);
extern int  rivatv_register_procfs (void);
extern void rivatv_unregister_procfs (void);

/*---------------------------------------------------------------------- 
 *
 * Modularization
 *
 *----------------------------------------------------------------------*/

#ifndef __stringify
#define __stringify_1(x)        #x
#define __stringify(x)          __stringify_1(x)
#endif

MODULE_AUTHOR	    ("Ferenc Bakonyi <fero@drama.obuda.kando.hu>");
MODULE_DESCRIPTION  ("Video4Linux driver for NVIDIA cards");
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 4, 10)
MODULE_LICENSE	    ("GPL");
#endif
MODULE_DEVICE_TABLE (pci, rivatv_pci_table);

MODULE_PARM	 (capbuffers, "1-" __stringify (RIVATV_MAX_CAPTURE_BUFFERS) "i");
MODULE_PARM_DESC (capbuffers, "Number of capture buffers (default = 2).");
MODULE_PARM	 (autoload, "i");
MODULE_PARM_DESC (autoload, "Turn decoder chip driver autoloading off (= 0) (On by default).");
MODULE_PARM      (card, "1-" __stringify (RIVATV_MAX) "i");
MODULE_PARM_DESC (card, "Specify card model, see CARDLIST file for a list.");
MODULE_PARM	 (debug, "i");
MODULE_PARM_DESC (debug, "Debug level: 0 = silent, 1 = verbose, 2 = very verbose (default = 1).");
MODULE_PARM	 (tvbox, "i");
MODULE_PARM_DESC (tvbox, "Indicate usage of the TV-Box (0 = off, 1 = PAL, 2 = NTSC-M) (Off by default).");
#ifdef CONFIG_MTRR
MODULE_PARM	 (mtrr, "i");
MODULE_PARM_DESC (mtrr, "Turn off (= 0) MTRR usage (default = 1).");
#endif

/* Definition of the video grabber device functionality. */

#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 5, 0)

struct __initdata video_device rivatv_video = { 
	name:		"rivatv video",
	type:		VID_TYPE_CAPTURE,
	hardware:	VID_HARDWARE_RIVA128,
	open:		rivatv_video_open,
	close:		rivatv_video_close,
	read:		NULL, // rivatv_video_read,
	write:		NULL, // rivatv_video_write,
	ioctl:		rivatv_video_ioctl,
	mmap:		rivatv_video_mmap,
};

#else

static struct file_operations rivatv_video_fops = {
        owner:          THIS_MODULE,
        open:           rivatv_video_open,
        release:        rivatv_video_close,
	read:		NULL, // rivatv_video_read,
	write:		NULL, // rivatv_video_write,
        mmap:           rivatv_video_mmap,
        ioctl:          rivatv_video_ioctl,
        llseek:         no_llseek,
};

static struct video_device rivatv_video = {
        owner:          THIS_MODULE,
        name:           "rivatv video",
        type:           VID_TYPE_CAPTURE,
        hardware:       VID_HARDWARE_RIVA128,
        fops:           &rivatv_video_fops,
};

#endif

/* Definition of the video teletext device. */

#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 5, 0)

struct __initdata video_device rivatv_vbi = { 
	name:		"rivatv vbi",
	type:		VID_TYPE_TELETEXT,
	hardware:	VID_HARDWARE_RIVA128,
	open:		rivatv_vbi_open,
	close:		rivatv_vbi_close,
	read:		NULL, // rivatv_vbi_read,
	write:		NULL, // rivatv_vbi_write,
	ioctl:		rivatv_vbi_ioctl,
};

#else

static struct file_operations rivatv_vbi_fops = {
        owner:          THIS_MODULE,
        open:           rivatv_vbi_open,
        release:        rivatv_vbi_close,
	read:		NULL, // rivatv_vbi_read,
	write:		NULL, // rivatv_vbi_write,
        ioctl:          rivatv_vbi_ioctl,
        llseek:         no_llseek,
};

static struct video_device rivatv_vbi = {
        owner:          THIS_MODULE,
        name:           "rivatv vbi",
        type:           VID_TYPE_TELETEXT,
        hardware:       VID_HARDWARE_RIVA128,
        fops:           &rivatv_vbi_fops,
};

#endif

/*---------------------------------------------------------------------- 
 *
 * Driver startup code
 *
 *----------------------------------------------------------------------*/

static struct __initdata rivatv_initdata driver_init = {
	video:	&rivatv_video,
	vbi:	&rivatv_vbi
};

#ifdef MODULE
int __init rivatv_init (void)
#else
int __init rivatv_setup (char *options)
#endif
{
	int err;

	PRINTK_INFO ("Video4Linux driver for NVIDIA cards\n");
	PRINTK_INFO ("Version 0.8.1\n");

	err = pci_module_init (&rivatv_driver);
	if (err)
		return err;
	err = rivatv_video_init ();
	if (err)
		return err;
#ifdef CONFIG_PROC_FS
	err = rivatv_register_procfs ();
	if (err)
		return err;
#endif

	return rivatv_register_v4l (&driver_init);
}

#ifdef MODULE
void __exit rivatv_cleanup (void) 
{
	rivatv_unregister_v4l ();
#ifdef CONFIG_PROC_FS
	rivatv_unregister_procfs ();
#endif
	cleanup_i2c_riva ();
	pci_unregister_driver (&rivatv_driver);
}

module_init (rivatv_init);
module_exit (rivatv_cleanup);
#endif /* MODULE */
