 /*
    tw98.c - generic i2c-driver for the
     
     - tw98 Analog to Digital Video Decoder
    
    Copyright (C) 2000 H. S. Magnuski <hankm@mtinet.com>

    Modifications to work with riva-tv - Steven Ellis <steven_ellis@users.sourceforge.net>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/* Module versioning */
#include <linux/config.h>
#if defined (CONFIG_MODVERSIONS) && !defined (MODVERSIONS)
#define MODVERSIONS
#endif
#include <linux/version.h>
#ifdef MODVERSIONS
#include <linux/modversions.h>
#endif
/* end module versioning */

#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/slab.h>
#include <linux/init.h>

#include <linux/i2c.h>
#include <linux/videodev.h>
#include <linux/video_decoder.h>

#include "tw98.h"

#define TW98_I2C_VERSION_ID 0xC0

#define TW98_DEBUG 1

#define TW98_PREFIX TW98_DEVICE_NAME ": "

#define TW98_PRINTK(format, args...) \
        printk (format , ## args)

#if TW98_DEBUG
#define TW98_DEBUG_PRINTK(format, args...) \
        TW98_PRINTK (format , ## args)
#else
#define TW98_DEBUG_PRINTK(format, args...)
#endif

#define	I2C_TW98		0x44

/* Possible TW98 addresses to scan, found only at 0x88 and/or 0x8a */
static unsigned short normal_i2c[] = { I2C_TW98, I2C_TW98 + 1, I2C_CLIENT_END };
static unsigned short normal_i2c_range[] = { I2C_CLIENT_END };

/* Magic definition of all other variables and things */
I2C_CLIENT_INSMOD;

/* Unique ID allocation */
static int tw98_id = 0;

/* For registering the I2C 'driver' driver */
static struct i2c_driver i2c_driver_tw98;

/* This functions sets partial bits of the register 'subaddress' from the
   i2c-client 'client'. Valid values of 'arg' may range between
   'range_begin' and 'range_end'; the corresponding bits to be set
   range from 'bit_begin' to 'bit_end' */
int tw98_set_register (struct i2c_client *client, int arg, int subaddress, 
		       int range_begin, int range_end,
		       int bit_begin, int bit_end) 
{
	s32 byte = 0x0;
	int i = 0, result = 0;

	/* be careful, check arguments :-) */
	if (arg < range_begin || arg > range_end)
		return -EINVAL;
	
	/* get the old register contents */
	if (-1 == (byte = i2c_smbus_read_byte_data (client, subaddress))) {
		TW98_PRINTK (TW98_PREFIX "FAILED reading register 0x%02x\n", subaddress);
		return -EFAULT;
	}

	/* clear out corresponding bits */
	for (i = bit_begin; i <= bit_end; i++)
		byte &= ~(1 << i);

	/* set these bits new */
	byte |= (arg << bit_begin);

	/* write register contents back */
	if (0 != (result = i2c_smbus_write_byte_data (client, subaddress, byte))) {
		TW98_PRINTK (TW98_PREFIX "FAILED writing register 0x%02x\n", subaddress);
		return -EFAULT;
	}
	
	/* everythings fine */
	return 0;
}

/* Read from a register */
int tw98_get_register (struct i2c_client *client, int subaddress)
{
	int val = i2c_smbus_read_byte_data (client, subaddress);

	if (val < 0) {
		TW98_PRINTK (TW98_PREFIX "FAILED reading register 0x%02x\n", subaddress);
		return -EFAULT;
	}
	return val;
}

/* This functions sets a range of 5 registers, used to switch
   from NTSC to PAL */
static void tw98_set_range (struct i2c_client *client, int subaddress, int value_0,
		     int value_1, int value_2, int value_3, int value_4)
{
	tw98_set_register (client, value_0, subaddress + 0, 0, 255, 0, 7);
	tw98_set_register (client, value_1, subaddress + 1, 0, 255, 0, 7);
	tw98_set_register (client, value_2, subaddress + 2, 0, 255, 0, 7);
	tw98_set_register (client, value_3, subaddress + 3, 0, 255, 0, 7);
	tw98_set_register (client, value_4, subaddress + 4, 0, 255, 0, 7);
}

#if TW98_DEBUG
/* Dump all registers */
static void tw98_dump_regs (struct i2c_client *client)
{
	int i, j;

	for (i = 0; i < TW98_NR_REGISTER; i += 16) {
		TW98_DEBUG_PRINTK ("0x%02x |", i);
		for (j = 0; j < 16 && i + j  < TW98_NR_REGISTER; j++) {
			TW98_DEBUG_PRINTK (" %02x", tw98_get_register (client, i + j));
		}
		TW98_DEBUG_PRINTK ("\n");
	}
}
#else
#define tw98_dump_regs(client)
#endif /* TW98_DEBUG */

/* Setup the scaler */
static void tw98_setup_scaler (struct i2c_client *client,
			       int src_x, int src_y, int src_width, int src_height,
			       int dst_width, int dst_height)
{
	int scale;

        TW98_DEBUG_PRINTK (" SCALE_IMAGE %d x %d", dst_width, dst_height);

	/* adjust vertical region */
	tw98_set_register (client, src_y & 0xff, TW98_VDELAY_LO, 0, 255, 0, 7);
	tw98_set_register (client, src_y >> 8,   TW98_CROP_HI,   0,   3, 6, 7);
	tw98_set_register (client, dst_height & 0xff, TW98_VACTIVE_LO, 0, 255, 0, 7);
	tw98_set_register (client, dst_height >> 8,   TW98_CROP_HI,    0,   3, 4, 5);
	scale = src_height * 256 / dst_height;
	tw98_set_register (client, scale & 0xff, TW98_VSCALE_LO, 0, 255, 0, 7);
	tw98_set_register (client, scale >> 8,   TW98_VSCALE_HI, 0,  15, 0, 4);
	
	/* adjust horizontal region */
	tw98_set_register (client, src_x & 0xff, TW98_HDELAY_LO, 0, 255, 0, 7);
	tw98_set_register (client, src_x >> 8,   TW98_CROP_HI,   0,   3, 2, 3);
	tw98_set_register (client, dst_width & 0xff, TW98_HACTIVE_LO, 0, 255, 0, 7);
	tw98_set_register (client, dst_width >> 8,   TW98_CROP_HI,    0,   3, 0, 1);
	scale = src_width * 256 / dst_width;
	tw98_set_register (client, scale & 0xff, TW98_HSCALE_LO, 0, 255, 0, 7);
	tw98_set_register (client, scale >> 8,   TW98_HSCALE_HI, 0, 255, 0, 7);
}

/* Interface to the world */
static int tw98_command (struct i2c_client *client, unsigned int cmd, void *arg)
{
	switch (cmd)
	{

	case DECODER_GET_CAPABILITIES:
	{
                struct video_decoder_capability *cap = arg;

                TW98_DEBUG_PRINTK ("command: DECODER_GET_CAPABILITIES\n");
		cap->flags
			= VIDEO_DECODER_PAL
                        | VIDEO_DECODER_SECAM
                        | VIDEO_DECODER_NTSC
                        | VIDEO_DECODER_AUTO
                        | VIDEO_DECODER_CCIR;
                cap->inputs = 4;
                cap->outputs = 0;
                break;
	}

	case DECODER_GET_STATUS:
	{
		int *iarg = arg;
                int res, reg;

		reg = tw98_get_register (client, TW98_CSTATUS);
                TW98_DEBUG_PRINTK ("command: DECODER_GET_STATUS = 0x%02x ->", reg);
                res = 0;

		if (reg & 0x80) {
			res |= DECODER_STATUS_GOOD;
			TW98_DEBUG_PRINTK (" GOOD");
		} else {
			TW98_DEBUG_PRINTK (" BAD");
		}

		reg = tw98_get_register (client, TW98_AUTODETECT);
		if (reg & 0x01) {
			TW98_DEBUG_PRINTK (" AUTODETECT");
		}
		reg = tw98_get_register (client, TW98_INFORM);
		if ((reg & 0x02) == 0x00) {
			res |= DECODER_STATUS_NTSC;
			TW98_DEBUG_PRINTK (" NTSC-M");
		} else if ((reg & 0x02) == 0x01) {
			res |= DECODER_STATUS_PAL;
			TW98_DEBUG_PRINTK (" PAL-B");
		}
		else {
			TW98_DEBUG_PRINTK (" <invalid PAL/NTSC>");
		}

		TW98_DEBUG_PRINTK ("\n");
		tw98_dump_regs (client);
		*iarg = res;
		break;
	}

	case DECODER_SET_NORM:
	{
                int *iarg = arg;
		int result;

                TW98_DEBUG_PRINTK ("command: DECODER_SET_NORM : ");

                switch (*iarg)
                {

                case VIDEO_MODE_NTSC:
                        TW98_DEBUG_PRINTK ("NTSC");
			result  = tw98_set_register (client, 0, TW98_AUTODETECT,
						     0,3,		/* valid values: 0 to 3 */
						     0,2);		/* bits to be used: 0-2 */
			result  = tw98_set_register (client, 0, TW98_INFORM,
						     0,1,		/* valid values: 0 to 1 */
						     0,1);		/* bits to be used: 0-1 */

			result  = tw98_set_register (client, 0x38, TW98_HSYNC,
						     0,255,		/* valid values: 0 to 255 */
						     0,7);		/* bits to be used: 0-7 */
			result  = tw98_set_register (client, 0xd4, TW98_LLENGTH,
						     0,255,		/* valid values: 0 to 255 */
						     0,7);		/* bits to be used: 0-7 */
			result  = tw98_set_register (client, 0x62, TW98_SYNCTRL,
						     0,255,		/* valid values: 0 to 255 */
						     0,7);		/* bits to be used: 0-7 */
			result  = tw98_set_register (client, 3, TW98_VFORMAT,
						     0,3,		/* valid values: 0 to 3 */
						     6,7);		/* bits to be used: 6-7 */

			/* Set Nominal, Maximum and Minimum TSC values */
			tw98_set_range (client, TW98_TSC1,  0x01, 0xeb, 0xae, 0xbf, 0x03);
			tw98_set_range (client, TW98_TSCM1, 0x01, 0xeb, 0xdc, 0x9f, 0x12);
			tw98_set_range (client, TW98_TSCL1, 0x01, 0xeb, 0x80, 0xde, 0xf8);
			tw98_setup_scaler (client, 0, 24, 766, 480, 720, 480);
                        break;

                case VIDEO_MODE_PAL:
                        TW98_DEBUG_PRINTK ("PAL");
			result  = tw98_set_register (client, 0, TW98_AUTODETECT,
						     0,3,		/* valid values: 0 to 3 */
						     0,2);		/* bits to be used: 0-2 */
			result  = tw98_set_register (client, 1, TW98_INFORM,
						     0,1,		/* valid values: 0 to 1 */
						     0,1);		/* bits to be used: 0-1 */

			result  = tw98_set_register (client, 0x28, TW98_HSYNC,
						     0,255,		/* valid values: 0 to 255 */
						     0,7);		/* bits to be used: 0-7 */
			result  = tw98_set_register (client, 0xe0, TW98_LLENGTH,
						     0,255,		/* valid values: 0 to 255 */
						     0,7);		/* bits to be used: 0-7 */
			result  = tw98_set_register (client, 0x00, TW98_SYNCTRL,
						     0,255,		/* valid values: 0 to 255 */
						     0,7);		/* bits to be used: 0-7 */
			result  = tw98_set_register (client, 0, TW98_VFORMAT,
						     0,3,		/* valid values: 0 to 3 */
						     6,7);		/* bits to be used: 6-7 */

			/* Set Nominal, Maximum and Minimum TSC values */
			tw98_set_range (client, TW98_TSC1,  0x01, 0x8c, 0xf7, 0x88, 0x5e);
			tw98_set_range (client, TW98_TSCM1, 0x01, 0x8d, 0x18, 0x4c, 0xfa);
			tw98_set_range (client, TW98_TSCL1, 0x01, 0x8c, 0xd6, 0xc3, 0xc3);
			tw98_setup_scaler (client, 0, 24, 946, 576, 704, 576);
                        break;

                case VIDEO_MODE_SECAM:
                        TW98_DEBUG_PRINTK ("SECAM");
			result  = tw98_set_register (client, 0, TW98_AUTODETECT,
						     0,3,		/* valid values: 0 to 3 */
						     0,2);		/* bits to be used: 0-2 */
			result  = tw98_set_register (client, 1, TW98_INFORM,
						     0,1,		/* valid values: 0 to 1 */
						     0,1);		/* bits to be used: 0-1 */
			tw98_setup_scaler (client, 0, 24, 946, 576, 704, 576);
                        break;

                case VIDEO_MODE_AUTO:
                        TW98_DEBUG_PRINTK ("AUTODETECT");
			result  = tw98_set_register (client, 3, TW98_AUTODETECT,
						     0,3,		/* valid values: 0 to 3 */
						     0,2);		/* bits to be used: 0-2 */
                        break;

                default:
                        TW98_DEBUG_PRINTK (" <invalid function>");
                        return -EINVAL;

                }
                TW98_DEBUG_PRINTK ("\n");
                break;
	}

	case DECODER_SET_INPUT:
	{
                int *iarg = arg;
		int result;

                TW98_DEBUG_PRINTK ("command: DECODER_SET_INPUT : %d", *iarg);

		result = tw98_set_register (client, *iarg, TW98_INFORM,
					    0,7,	/* valid values: 0-7 */
					    2,4);	/* bits to be used: 2-4 */

                /* set S-Video Mode */
		if (*iarg >= 4) {
			result = tw98_set_register (client, 1, TW98_INFORM,
						    0,1,	/* valid values: 0-1 */
						    4,4);	/* bit to be used: 4 */
                        TW98_DEBUG_PRINTK (" S-Video");
                } else {
			result = tw98_set_register (client, 0, TW98_INFORM,
						    0,1,	/* valid values: 0-1 */
						    4,4);	/* bit to be used: 4 */
                        TW98_DEBUG_PRINTK (" CVBS");
                }
                TW98_DEBUG_PRINTK ("\n");
                break;
	}

	case DECODER_SET_OUTPUT:
	{
                int *iarg = arg;

                TW98_DEBUG_PRINTK ("command: DECODER_SET_OUTPUT : %d\n", *iarg);

                /* not much choice of outputs */
                if (*iarg != 0) {
                        return -EINVAL;
                }
                break;
	}

	case DECODER_ENABLE_OUTPUT:
	{
                int *iarg = arg;
		int result;

                TW98_DEBUG_PRINTK ("command: DECODER_ENABLE_OUTPUT: %d\n", *iarg);

		if (*iarg) {
			result  = tw98_set_register (client, 0xa2, TW98_OPFORM,
						     0,255,		/* valid values: 0 to 255 */
						     0,7);		/* bits to be used: 0-7 */
		} else {
			result  = tw98_set_register (client, 0xa6, TW98_OPFORM,
						     0,255,		/* valid values: 0 to 255 */
						     0,7);		/* bits to be used: 0-7 */
		}
		break;
	}

	case DECODER_SET_PICTURE:
	{
                struct video_picture *pic = arg;
		int result;

                TW98_DEBUG_PRINTK ("command: DECODER_SET_PICTURE\n");

		result  = tw98_set_register (client, (pic->brightness - 32768) >> 8, TW98_BRIGHTNESS,
					     0,255,		/* valid values: 0 to 255 */
					     0,7);		/* bits to be used: 0-7 */

		result  = tw98_set_register (client, pic->contrast >> 10, TW98_CONTRAST,
					     0,63,		/* valid values: 0 to 63 */
					     0,5);		/* bits to be used: 0-5 */

		result  = tw98_set_register (client, pic->colour >> 8, TW98_SAT_U,
					     0,255,		/* valid values: 0 to 255 */
					     0,7);		/* bits to be used: 0-7 */

		result  = tw98_set_register (client, pic->colour >> 8, TW98_SAT_V,
					     0,255,		/* valid values: 0 to 255 */
					     0,7);		/* bits to be used: 0-7 */

		result  = tw98_set_register (client, pic->hue >> 10, TW98_HUE,
					     0,63,		/* valid values: 0 to 63 */
					     0,5);		/* bits to be used: 0-5 */
                break;
	}

#ifdef DECODER_DUMP
	case DECODER_DUMP:
                TW98_DEBUG_PRINTK ("command: DECODER_DUMP\n");
		tw98_dump_regs (client);
		break;
#endif

#ifdef GET_NR_OF_REGISTERS
        case GET_NR_OF_REGISTERS:
                TW98_DEBUG_PRINTK ("command: NUM_REGISTERS = %d\n", TW98_NR_REGISTER);
                return TW98_NR_REGISTER;
#endif

	default:
                TW98_DEBUG_PRINTK ("command: <invalid> = 0x%x\n", cmd);
		return -EINVAL;
	}
	return 0;
}

/* This function initalizes a TW98 to useful values */
static int tw98_initialize (struct i2c_client* client)
{
	int 	i = 0, result = 0;

#define RESET 32
	u8	s[RESET][2] = {	{ TW98_ACNTL,		0x80 }, /* 0x06 */
				{ TW98_INFORM,		0x61 }, /* 0x02 */
				{ TW98_OPFORM,		0xa2 }, /* 0x03 */ /* 0x86 for BRS direct */
				{ TW98_HSLEN,		0x40 }, /* 0x04 */
				{ TW98_POLARITY,	0x00 }, /* 0x05 */
				{ TW98_CROP_HI,		0x22 }, /* 0x07 */
				{ TW98_VDELAY_LO,	0x18 }, /* 0x08 */
				{ TW98_VACTIVE_LO,	0x40 }, /* 0x09 */
				{ TW98_HDELAY_LO,	0x00 }, /* 0x0a */
				{ TW98_HACTIVE_LO,	0xc0 }, /* 0x0b */
				{ TW98_VSCALE_HI,	0x01 }, /* 0x0c */
				{ TW98_VSCALE_LO,	0x00 }, /* 0x0d */
				{ TW98_HSCALE_HI,	0x01 }, /* 0x0e */
				{ TW98_HSCALE_LO,	0x58 }, /* 0x0f */
				{ TW98_BRIGHTNESS,	0xca }, /* 0x10 */
				{ TW98_CONTRAST,	0x2f }, /* 0x11 */
				{ TW98_SHARPNESS,	0x33 }, /* 0x12 */
				{ TW98_SAT_U,		0x71 }, /* 0x13 */
				{ TW98_SAT_V,		0x50 }, /* 0x14 */
				{ TW98_HUE,		0x20 }, /* 0x15 */
				{ TW98_CKILL,		0xd0 }, /* 0x16 */
				{ TW98_CORING,		0x02 }, /* 0x17 */
				{ TW98_VBICNTL,		0x18 }, /* 0x19 */
				{ TW98_CC_STATUS,	0xc0 }, /* 0x1a */
				{ TW98_AGCGAIN,		0x21 }, /* 0x1c */
				{ TW98_GPIO,		0xf0 }, /* 0x1d */
				{ TW98_NOISE,           0x05 }, /* 0x1f */
#if 1
				{ TW98_SYNCL,		0x14 }, /* 0xd3 */
				{ TW98_SHARP1,		0x18 }, /* 0xd8 */
				{ TW98_LLENGTH,		0xb4 }, /* 0xdf */
				{ TW98_HSK,		0x04 }, /* 0xe2 */
				{ TW98_SYNCTRL,		0x62 }  /* 0xe6 */
#endif
				};

	/* initial registers */
	TW98_DEBUG_PRINTK (TW98_PREFIX "initializing\n");
	for (i = 0; i < RESET; i++) {
		if (0 != (result = i2c_smbus_write_byte_data (client, s[i][0], s[i][1]))) {
			TW98_PRINTK (TW98_PREFIX "init error - reset on %d\n", i);
			return -EFAULT;
		}
	}
	return 0;
}

/* This function is called by i2c_probe */
int tw98_detect (struct i2c_adapter *adapter, int address, unsigned short flags, int kind)
{
	struct i2c_client *client;
	int err = 0, version;

	/* let's see whether this adapter can support what we need */
	if (!i2c_check_functionality (adapter, I2C_FUNC_SMBUS_READ_BYTE_DATA |
				      I2C_FUNC_SMBUS_WRITE_BYTE_DATA))
		return 0;

	/* allocate memory for client structure */
	client = kmalloc (sizeof (struct i2c_client), GFP_KERNEL);
        if (NULL == client)
		return -ENOMEM;
	
	/* fill client structure */
	strcpy (client->name, TW98_DEVICE_NAME);
	client->id = tw98_id++;
	client->flags = 0;
	client->addr = address;
	client->adapter = adapter;
	client->driver = &i2c_driver_tw98;
	client->data = NULL;

	/* chip detection */
	if (kind < 0) {
		/* Get chip version */
		version = i2c_smbus_read_byte_data (client, TW98_ID);

		/* If we could not get version, bail out */
		if (version == -1) {
			kfree (client);
			return 0;
		}

		/* If not a TW98, bail out */
		if ((version & 0xf8) != TW98_I2C_VERSION_ID) {
			TW98_PRINTK (TW98_PREFIX "Unknown TWxx chip @ 0x%02x found."
				     " (Maybe TW99) Chip version: %#x\n", 2 * address, version);
			kfree (client);
			return 0;
		}

		/* Chip IS a TW98, log into syslog  */
		TW98_PRINTK (TW98_PREFIX "video decoder chip found."
			     " Chip version: %#x (rev. %d)\n", version >> 3, version & 0x07);
	}
	else {
		TW98_PRINTK (TW98_PREFIX "detection skipped\n");
	}

	/* initialize registers */
	if (0 != tw98_initialize (client)) {
		TW98_PRINTK (TW98_PREFIX "could not initialize chipset, continuing...\n");
	}
	
	/* tell the i2c layer a client has arrived */
	if (0 != (err = i2c_attach_client (client))) {
		TW98_PRINTK (TW98_PREFIX "FAILED attaching client\n");
		kfree (client);
		return err;
	}

	TW98_PRINTK (TW98_PREFIX "detected @ 0x%02x on adapter %s\n", 2 * address, client->adapter->name);
	return 0;
}

/* Called when a new I2C bus found */
static int tw98_attach (struct i2c_adapter *adapter)
{
	return i2c_probe (adapter, &addr_data, &tw98_detect);
}

/* Called on exit */
static int tw98_detach (struct i2c_client *client)
{
	int err = 0;

	if (0 != (err = i2c_detach_client (client))) {
		TW98_PRINTK (TW98_PREFIX "Client deregistration failed, client not detached.\n");
		return err;
	}
	
	kfree (client);
	return 0;
}

void tw98_inc_use (struct i2c_client *client)
{
	MOD_INC_USE_COUNT;
}

void tw98_dec_use (struct i2c_client *client)
{
	MOD_DEC_USE_COUNT;
}

static struct i2c_driver i2c_driver_tw98 = {
	"TW98",
	I2C_DRIVERID_TW98,
	I2C_DF_NOTIFY,
	tw98_attach,
	tw98_detach,
	tw98_command,
	tw98_inc_use,
	tw98_dec_use
};

/*---------------------------------------------------------------------- 
 *      Modularization
 *----------------------------------------------------------------------*/

MODULE_AUTHOR("Hank Magnuski <hankm@mtinet.com>");
MODULE_DESCRIPTION("Techwell 98 driver");
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 4, 10)
MODULE_LICENSE("GPL");
#endif

int __init tw98_init (void)
{
	int res = 0;

	if (0 != (res = i2c_add_driver (&i2c_driver_tw98))) {
		TW98_PRINTK (TW98_PREFIX "Driver registration failed, module not inserted.\n");
		return res;
	}
	return 0;
}

void __init tw98_cleanup (void)
{	
	int res = 0;

	if (0 != (res = i2c_del_driver (&i2c_driver_tw98))) {
		TW98_PRINTK (TW98_PREFIX "Driver deregistration failed, module not removed.\n");
	}
}

module_init (tw98_init);
module_exit (tw98_cleanup);
