/*
    vpx32xx.c  MICRONAS INTERMETALL VPX32XX video pixel decoder driver

    (c) 1999-2001 Peter Schlaf <peter.schlaf@org.chemie.uni-giessen.de>
        project started by Jens Seidel <jens.seidel@hrz.tu-chemnitz.de>
                   
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

/* module versioning */
#include <linux/config.h>
#if defined (CONFIG_MODVERSIONS) && !defined (MODVERSIONS)
# define MODVERSIONS
#endif
#include <linux/version.h>
#ifdef MODVERSIONS
#include <linux/modversions.h>
#endif
/* end module versioning */

#include <linux/module.h>
#include <linux/types.h>
#include <linux/delay.h>
#include <linux/errno.h>
#include <linux/slab.h>
#include <linux/videodev.h>
#include <linux/video_decoder.h>
#include <linux/i2c.h>
#include <linux/init.h>

#define VPX32XX_DEBUG 1

#if VPX32XX_DEBUG
#define DPRINTK(format, args...) printk (format , ## args)
#else
#define DPRINTK(format, args...)
#endif

#ifndef I2C_VPX32XX
#define I2C_VPX32XX	(0x86 / 2)	/* i2c_address / 2 */
#endif

#ifndef I2C_DRIVERID_VPX32XX
#define I2C_DRIVERID_VPX32XX	42      /* video decoder+vbi/vtxt	*/
#endif

#define VPX32XX_MAX_INPUT	8	/* 3 CVBS + 5 SVHS modes are possible */
#define VPX32XX_MAX_OUTPUT	0	/* these chips can only decode */

struct vpx32xx {
	int	part_number;
#define VPX3220A 0x4680      /* 16-bit part numbers */
#define VPX3216B 0x4260
#define VPX3214C 0x4280
#define VPX3225D 0x7230
#define VPX3224D 0x7231

	unsigned char	fprd;
	unsigned char	fpwr;
	unsigned char	fpdat;
	unsigned char	fpsta;

	int	norm;
#define VPX_PALBGHI	0x00
#define VPX_PALM	0x04
#define VPX_PALN	0x05
#define VPX_PAL60	0x06
#define VPX_NTSCM	0x01
#define VPX_NTSC44	0x03
#define VPX_NTSCCOMB	0x07
#define VPX_SECAM	0x02
	int	input;

	__u16	hue;
	__u16	saturation;
	__u16	brightness1;
	__u16	brightness2;
	__u16	contrast1;
	__u16	contrast2;
};

static int id = 0;

static unsigned short int normal_i2c[] = { I2C_VPX32XX, I2C_CLIENT_END };
static unsigned short int normal_i2c_range[] = { I2C_CLIENT_END };

I2C_CLIENT_INSMOD;

static struct i2c_driver driver;

//*****************************************************************************/

/* Write one byte to i2c chip register */
static int vpx32xx_writeI2C (struct i2c_client *client, unsigned char subaddr, unsigned char data)
{
	unsigned char buffer[2] = { subaddr, data };

	if (sizeof (buffer) != i2c_master_send (client, buffer, sizeof (buffer))) {
		printk (KERN_WARNING "vpx32xx.o: vpx32xx_writeI2C FAILED! (I/O error)\n");
		return -EIO;
	}
	DPRINTK (KERN_INFO "vpx32xx.o: vpx32xx_writeI2C (subaddr=0x%02x, data=0x%02x) OK\n", subaddr, data);
	return 0;
}

/* Read and return one byte from i2c chip register */
static unsigned char vpx32xx_readI2C (struct i2c_client *client, unsigned char subaddr)
{
	unsigned char read;
	struct i2c_msg msg[] = {
		{ client->addr, 0,        1, &subaddr },
		{ client->addr, I2C_M_RD, 1, &read    }
	};
	int size = sizeof (msg) / sizeof (msg[0]); 
	
	if (size != i2c_transfer (client->adapter, msg, size)) {
		printk (KERN_WARNING "vpx32xx.o: vpx32xx_readI2C FAILED! (I/O error)\n");
		return -EIO;
	}
	DPRINTK (KERN_INFO "vpx32xx.o: vpx32xx_readI2C (subaddr=0x%02x, read=0x%02x) OK\n", subaddr, read);
	return read;
}

/*******************************/

/* Write to vpx FP-RAM register */
static int vpx32xx_writeFP (struct i2c_client *client, int subaddr, int data)
{
	struct vpx32xx *decoder = client->data;

	unsigned char load[3];
	unsigned char write = decoder->fpsta;
	unsigned char read;
	struct i2c_msg msg[2] = {
		{ client->addr, 0,        1, &write },
		{ client->addr, I2C_M_RD, 1, &read  }
	};
	int size = sizeof (msg) / sizeof (msg[0]); 
	int i;

	load[0] = decoder->fpwr;
	load[1] = (unsigned char) (subaddr >> 8);
	load[2] = (unsigned char) (subaddr & 0x00FF);

	if (sizeof (load) != i2c_master_send (client, load, sizeof (load))) {
		printk (KERN_WARNING "vpx32xx.o: vpx32xx_writeFP FAILED! (I/O error) [1]\n");
		return -EIO;
	}
	
	load[0] = decoder->fpdat;
	load[1] = (unsigned char) (data >> 8);
	load[2] = (unsigned char) (data & 0x00FF);

	for (i = 0; i < 10; i++) {	/* FP busy ? */
		if (size != i2c_transfer (client->adapter, msg, size)) {
			printk (KERN_WARNING "vpx32xx.o: vpx32xx_writeFP FAILED! (I/O error) [2]\n");
			return -EIO;
		}
							
		if (!(read & 0x04)) {
			if (sizeof (load) != i2c_master_send (client, load, sizeof (load))) {
				printk (KERN_WARNING "vpx32xx.o: vpx32xx_writeFP FAILED! (I/O error) [3]\n");
				return -EIO;
			}
			DPRINTK (KERN_INFO "vpx32xx.o: vpx32xx_writeFP (subaddr=0x%04x, data=0x%04x) OK\n", subaddr, data);
			return 0;
		}
	}
	printk (KERN_WARNING "vpx32xx.o: vpx32xx_writeFP: FP processor is busy!\n");
	return -EBUSY;
}

/* Read from vpx FP-RAM register */
static int vpx32xx_readFP (struct i2c_client *client, int subaddr)
{
	struct vpx32xx *decoder = client->data;

	unsigned char load[3];
	unsigned char write = decoder->fpsta;
	unsigned char read[2];
	struct i2c_msg msg[2] = {
		{ client->addr, 0,        1, &write },
		{ client->addr, I2C_M_RD, 1, read   }
	};
	int size = sizeof (msg) / sizeof (msg[0]); 
	int i;
	
	load[0] = decoder->fprd;
	load[1] = (unsigned char) (subaddr >> 8);
	load[2] = (unsigned char) (subaddr & 0x00FF);

	if (sizeof (load) != i2c_master_send (client, load, sizeof (load))) {
		printk (KERN_WARNING "vpx32xx.o: vpx32xx_readFP FAILED! (I/O error) [1]\n");
		return -EIO;
	}
	
	for (i = 0; i < 10; i++) {	/* FP busy ? */
		msg[1].len = 1;
		write = decoder->fpsta;
		if (size != i2c_transfer (client->adapter, msg, size)) {
			printk (KERN_WARNING "vpx32xx.o: vpx32xx_readFP FAILED! (I/O error) [2]\n");
			return -EIO;
		}
							
		if (!(read[0] & 0x04)) {
			msg[1].len = 2;
			write = decoder->fpdat;

			if (size != i2c_transfer (client->adapter, msg, size)) {
				printk (KERN_WARNING "vpx32xx.o: vpx32xx_readFP FAILED! (I/O error) [3]\n");
				return -EIO;
			}
			DPRINTK (KERN_INFO "vpx32xx.o: vpx32xx_readFP (subaddr=0x%04x, read=0x%04x) OK\n", subaddr, ((read[0] << 8) | read[1]));
			return ((read[0] << 8) | read[1]);
		}
	}
	printk (KERN_WARNING "vpx32xx.o: vpx32xx_readFP: FP processor is busy!\n");
	return -EBUSY;
}

/*****************************************************************************/

/* Set picture for VPX3214, VPX3216 or VPX3220 */
static int vpx32xx_set_picture (struct i2c_client *client)
{
	struct vpx32xx *decoder = client->data;
	int ret = 0;
	
	/* set brightness */
	ret |= vpx32xx_writeI2C (client, 0xE6, ((decoder->brightness1 - 32768) >> 8) & 0xFF);

	/* set contrast */
	ret |= vpx32xx_writeI2C (client, 0xE7, ((decoder->contrast1 >> 10) & 0x003F));

	/* set saturation */
	ret |= vpx32xx_writeFP (client, 0xA0, (decoder->saturation >> 4) | 1);

	/* set hue (for NTSC only) */
	ret |= vpx32xx_writeFP (client, 0x1C, ((decoder->hue - 32768) >> 6) & 0x03FF); /* ntsc tint angle: +-512 */

	return ret;
}

/* Set picture for VPX3224 or VPX3225 */
static int vpx322x_set_picture (struct i2c_client *client)
{
	struct vpx32xx *decoder = client->data;
	int ret = 0;
	
	/* set brightness */
	ret |= vpx32xx_writeFP (client, 0x127, ((decoder->brightness1 - 32768) >> 8) & 0xFF);
	ret |= vpx32xx_writeFP (client, 0x131, ((decoder->brightness2 - 32768) >> 8) & 0xFF);

	/* set contrast */
	ret |= vpx32xx_writeFP (client, 0x128, ((decoder->contrast1 >> 10) | 0x0100)); /* set clamping level=16 */
	ret |= vpx32xx_writeFP (client, 0x132, ((decoder->contrast2 >> 10) | 0x0100)); /* set clamping level=16 */

	/* set saturation */
	ret |= vpx32xx_writeFP (client, 0x30, (decoder->saturation >> 4) | 1);

	/* set hue (for NTSC only) */
	ret |= vpx32xx_writeFP (client, 0xDC, ((decoder->hue - 32768) >> 6) & 0x03FF); /* ntsc tint angle: +-512 */

	/* latching */
	ret |= vpx32xx_writeFP (client, 0x140, 0x860);	/* set latch Timing mode and WinTable #1 + #2 */
	
	return ret;
}

/*****************************************************************************/

/* Set video_norm for VPX3214, VPX3216 or VPX3220 */
static int vpx32xx_set_video_norm (struct i2c_client *client, int video_norm)
{
	struct vpx32xx *decoder = client->data;
	int i, choice, ret = 0;

	const int init0[][2] = {  /* FP init */
	/* addr    value                     */
	{ 0x00A0, 0x0001 },	/*          */
	{ 0x008F, 0x0000 },	/*          */
	{ 0x004B, 0x0298 },	/*          */
	{ 0x001C, 0x01F4 }	/*          */
	}; 

	const int init1[][2] = {  /* I2C init */
	/* addr    value                      */
	{ 0x00D8, 0x00A6 },	/* old: 0x00A4 */
	{ 0x00E8, 0x00F8 },	/*            */
	{ 0x00EA, 0x0010 },	/*            */
	{ 0x00F0, 0x000B },	/*            */
	{ 0x00F1, 0x008C },	/*            */
	{ 0x00F2, 0x001B },	/*            */
	{ 0x00F8, 0x0002 },	/*            */
	{ 0x00F9, 0x0025 }	/*            */
	}; 

	const int norms[3][11] = {
	/*                  88      89      8a      8b      8c      8d      f0      f2      c1      e7      a0    */
	/* PAL   */	{ 0x0007, 0x0130, 0x0130, 0x0000, 0x0330, 0x0330, 0x0077, 0x0011, 0x0058, 0x026E, 0x0B32 },
	/* NTSC  */	{ 0x000A, 0x00F0, 0x00F0, 0x0000, 0x0290, 0x02B0, 0x0077, 0x0013, 0x0060, 0x020A, 0x0B32 },
	/* SECAM */	{ 0x0007, 0x0130, 0x0130, 0x0000, 0x0330, 0x0330, 0x0077, 0x0015, 0x006C, 0x026E, 0x0219 } };

	switch (video_norm) {
		case VIDEO_MODE_PAL: 	choice = 0; break;	/* PAL B,G,H,I 625/50	*/
		case VIDEO_MODE_NTSC: 	choice = 1; break;	/* NTSC-M 525/60 	*/
		case VIDEO_MODE_SECAM: 	choice = 2; break;	/* SECAM 625/50 	*/
		default:
			printk (KERN_WARNING "vpx32xx.o: vpx32xx_set_video_norm INVALID! (norm=%d)\n", video_norm);
			return -EINVAL;
	}

	if (decoder->part_number == VPX3214C)
		ret |= vpx32xx_writeI2C (client, 0xAA, 0x00);	/* normal power mode */

	/*	FP init0 	*/
	for (i = 0; i < (sizeof (init0) / (2 * sizeof (int))); i++)
		ret |= vpx32xx_writeFP (client, init0[i][0], init0[i][1]);

	/*	I2C init1 	*/
	for (i = 0; i < (sizeof (init1) / (2 * sizeof (int))); i++)
		ret |= vpx32xx_writeI2C (client, (unsigned char) (init1[i][0]), (unsigned char) (init1[i][1]));

	ret |= vpx32xx_writeFP (client, 0x0088, norms[choice][0]); 
	ret |= vpx32xx_writeFP (client, 0x0089, norms[choice][1]);
	ret |= vpx32xx_writeFP (client, 0x008A, norms[choice][2]);
	ret |= vpx32xx_writeFP (client, 0x008B, norms[choice][3]); 
	ret |= vpx32xx_writeFP (client, 0x008C, norms[choice][4]); 
	ret |= vpx32xx_writeFP (client, 0x008D, norms[choice][5]); 

	ret |= vpx32xx_writeFP (client, 0x00F0, norms[choice][6]); 
	ret |= vpx32xx_writeFP (client, 0x00F2, norms[choice][7]); 
	ret |= vpx32xx_writeFP (client, 0x00C1, norms[choice][8]); 
	ret |= vpx32xx_writeFP (client, 0x00E7, norms[choice][9]); 
	ret |= vpx32xx_writeFP (client, 0x00A0, norms[choice][10]); 

	if (0 == ret)
		decoder->norm = video_norm;
	return ret;
}

/* Set video_norm for VPX3224 and VPX3225 */
static int vpx322x_set_video_norm (struct i2c_client *client, __u16 video_norm)
{
	struct vpx32xx *decoder = client->data;
	int ret = 0;
	const int norms[8][11] = {
	/*                ver.b.  ver.li. ver.lo. hor.b.  hor.l.  nr.px.  sel.p/c st.v.a. e.v.a.  col.pr. */
	/* PAL       */ { 0x0007, 0x0130, 0x0130, 0x0000, 0x0330, 0x0330, 0x0100, 0x0008, 0x0330, 0x07FF, 0x0000 }, 
	/* NTSC      */ { 0x000A, 0x00F0, 0x00F0, 0x0000, 0x0290, 0x02B0, 0x0100, 0x0030, 0x02A8, 0x07FF, 0x0001 },
	/* SECAM     */ { 0x0007, 0x0130, 0x0130, 0x0000, 0x0330, 0x0330, 0x0100, 0x0008, 0x0330, 0x017F, 0x0002 },

	/* PAL 60    */ { 0x000A, 0x00F0, 0x00F0, 0x0000, 0x0290, 0x02B0, 0x0100, 0x0030, 0x02A8, 0x07FF, 0x0006 },
	/* PAL M     */ { 0x000A, 0x00F0, 0x00F0, 0x0000, 0x0290, 0x02B0, 0x0100, 0x0030, 0x02A8, 0x07FF, 0x0004 },

	/* PAL N     */ { 0x0007, 0x0130, 0x0130, 0x0000, 0x0330, 0x0330, 0x0100, 0x0008, 0x0330, 0x07FF, 0x0005 }, 
	/* NTSC44    */ { 0x000A, 0x00F0, 0x00F0, 0x0000, 0x0290, 0x02B0, 0x0100, 0x0030, 0x02A8, 0x07FF, 0x0003 },
	/* NTSC COMP */ { 0x000A, 0x00F0, 0x00F0, 0x0000, 0x0290, 0x02B0, 0x0100, 0x0030, 0x02A8, 0x07FF, 0x0007 } };

	if (video_norm > 7) {
		printk (KERN_WARNING "vpx32xx.o: vpx32xx_set_video_norm INVALID! (norm=%d)\n", video_norm);
		return -EINVAL;
	}
	
	/* Output */
	ret |= vpx32xx_writeI2C (client, 0xAA, 0x00);		/* Low Power Mode, bit 4 on TDO-pin and 0 */
	ret |= vpx32xx_writeI2C (client, 0xF2, 0x0F);		/* Output Enable */

	/* Formatter */
	ret |= vpx32xx_writeFP (client, 0x0150, 0x0010);	/* Format Selection */

	/* Output Multiplexer */
	ret |= vpx32xx_writeFP (client, 0x0154, 0x0000);	/* Multiplexer/Multipurpose output */ 

	/* Load Table for Window #1 */
	ret |= vpx32xx_writeFP (client, 0x0120, norms[video_norm][0]);  /* Vertical Begin                */
	ret |= vpx32xx_writeFP (client, 0x0121, norms[video_norm][1]);  /* Vertical Lines In             */
	ret |= vpx32xx_writeFP (client, 0x0122, norms[video_norm][2]);  /* Vertical Lines Out            */
	ret |= vpx32xx_writeFP (client, 0x0123, norms[video_norm][3]);  /* Horizontal Begin              */
	ret |= vpx32xx_writeFP (client, 0x0124, norms[video_norm][4]);  /* Horizontal Length             */
	ret |= vpx32xx_writeFP (client, 0x0125, norms[video_norm][5]);  /* Number of Pixels              */
	ret |= vpx32xx_writeFP (client, 0x0126, norms[video_norm][6]);  /* Selection for peaking/coring  */

	/* Load Table for Window #2 */
	ret |= vpx32xx_writeFP (client, 0x012B, 0x0C00);                /* Vertical Lines In (window #2 disabled) */

	/* HVREF */
	ret |= vpx32xx_writeFP (client, 0x0151, norms[video_norm][7]);  /* Startpos. of programmable video aktive */
	ret |= vpx32xx_writeFP (client, 0x0152, norms[video_norm][8]);  /* Endpos. of programmable video aktive   */
	ret |= vpx32xx_writeFP (client, 0x0153, 0x0027);                /* Length/Polarity of HREF, VREF, FIELD   */

	/* Standard Selection */	
	ret |= vpx32xx_writeFP (client, 0x0020, norms[video_norm][10]); /* tv norm setting */

	/* Load Table for VBI-Window */
	ret |= vpx32xx_writeFP (client, 0x0138, 0x0800);                /* Control VBI-Window (vbi window disabled) */ 

	/* Control Word */
	ret |= vpx32xx_writeFP (client, 0x0140, 0x0864);                /* Latch for Window #1 + #2 */ 

	/* Color Processing */	
	ret |= vpx32xx_writeFP (client, 0x0030, norms[video_norm][9]);  /* acc reference setting */

	if (0 == ret)
		decoder->norm = video_norm;
	return ret;
}

/* Input select for VPX3214, VPX3216, VPX3218, VPX3224 and VPX3225 */
static int vpx32xx_selmux (struct i2c_client *client, int video_input)
{
	struct vpx32xx *decoder = client->data;
	int data, ret = 0;
	const int input_vpx322x[VPX32XX_MAX_INPUT][2] = {
	/*  0x21 , s-vhs                              */
	{  0x0224,   0 },   /* CVBS (VIN3)            */
	{  0x0225,   0 },   /* CVBS (VIN2)            */
	{  0x0226,   0 },   /* CVBS (VIN1)            */

	{  0x0224,   1 },   /* S-VHS (Y-VIN3, C-CIN)  */
	{  0x0225,   1 },   /* S-VHS (Y-VIN2, C-CIN)  */
	{  0x0226,   1 },   /* S-VHS (Y-VIN1, C-CIN)  */
	{  0x0220,   1 },   /* S-VHS (Y-VIN3, C-VIN1) */
	{  0x0221,   1 }};  /* S-VHS (Y-VIN2, C-VIN1) */

	const int input_vpx32xx[VPX32XX_MAX_INPUT][2] = {
	/* 0x33	 , s-vhs                              */
	{  0x0000,   0x11 },  /* CVBS (VIN3)            */
	{  0x0001,   0x11 },  /* CVBS (VIN2)            */
	{  0x0002,   0x11 },  /* CVBS (VIN1)            */

	{  0x000C,   0x31 },   /* S-VHS (Y-VIN3, C-CIN)  */
	{  0x000D,   0x31 },   /* S-VHS (Y-VIN2, C-CIN)  */
	{  0x000E,   0x31 },   /* S-VHS (Y-VIN1, C-CIN)  */
	{  0x0008,   0x31 },   /* S-VHS (Y-VIN3, C-VIN1) */
	{  0x0009,   0x31 }};  /* S-VHS (Y-VIN2, C-VIN1) */

	switch (decoder->part_number) {
		case VPX3224D: case VPX3225D:
			data = vpx32xx_readFP (client, 0x20) & ~(0x0040);	/* read word and erase bit 6 */
			if (0 > data)
				return data; 
			ret |= vpx32xx_writeFP (client, 0x20, (data | (input_vpx322x[video_input][1] << 6)));
			ret |= vpx32xx_writeFP (client, 0x21, input_vpx322x[video_input][0]);
			break;
		case VPX3220A: case VPX3216B: case VPX3214C:
			ret |= vpx32xx_writeI2C (client, 0x33, (unsigned char) (input_vpx32xx[video_input][0]));
			ret |= vpx32xx_writeFP (client, 0xF2, input_vpx32xx[video_input][1]);
			break;
		default:
			return -EINVAL;
	}
	if (0 == ret)
		decoder->input = video_input;
	return ret;
}

/* Dump VPX3224 or VPX3225 registers */
static int vpx322x_dump (struct i2c_client *client)
{
	int i, data_0, data_1, ret = 0;
	 
	printk (KERN_INFO "vpx32xx.o: vpx32xx dump i2c registers:\n");
	for (i = 0; i < 0x100; i = i + 2) {
		ret |= data_0 = vpx32xx_readI2C (client, i);
		ret |= data_1 = vpx32xx_readI2C (client, i + 1);
		printk (KERN_INFO "vpx32xx.o: vpx32xx_i2c_adr: 0x%02x = 0x%02x,  0x%02x = 0x%02x\n", i, data_0, i + 1, data_1);
	}

	printk (KERN_INFO "vpx32xx.o: vpx32xx dump fp-ram registers:\n");
	for (i = 0; i < 0x180; i = i + 2) {
		ret |= data_0 = vpx32xx_readFP (client, i);
		ret |= data_1 = vpx32xx_readFP (client, i + 1);
		printk(KERN_INFO "vpx32xx.o: vpx32xx_fp_adr: 0x%03x = 0x%03x,  0x%03x = 0x%03x\n", i, data_0, i + 1, data_1);
	}
	printk (KERN_INFO "vpx32xx.o: end of vpx32xx register dump\n");
	return ret;
}

/*****************************************************************************/

#if 0
/* VBI sliced data only possible on VPX3225 */
static int vpx3225_init_vbi_sliced (struct i2c_client *client)
{
	int ret = 0;

	/*	start vbi-initialisation	*/
	/*                              addr   value                                                            */
	ret |= vpx32xx_writeFP (client, 0x0121, 0x0C00); /* Vertical Lines (window #1 disable)                   */

	ret |= vpx32xx_writeI2C (client, 0xC7, 0x01);    /* accumulator mode: reset DC, AC, and FLT accu         */
	ret |= vpx32xx_writeI2C (client, 0xC7, 0x30);    /* accumulator mode: dc and ac adaptation enable        */
	ret |= vpx32xx_writeI2C (client, 0xCE, 0x15);    /* bit error tolerance                                  */
	ret |= vpx32xx_writeI2C (client, 0xB8, 0x00);    /* clock run-in and framing code dont care mask high   */
	ret |= vpx32xx_writeI2C (client, 0xB9, 0x00);    /* clock run-in and framing code dont care mask mid    */
	ret |= vpx32xx_writeI2C (client, 0xBA, 0x03);    /* clock run-in and framing code dont care mask low    */ 
	ret |= vpx32xx_writeI2C (client, 0xC8, 0x40);    /* sync slicer                                          */
	ret |= vpx32xx_writeI2C (client, 0xC0, 0x10);    /* soft slicer                                          */ 
	ret |= vpx32xx_writeI2C (client, 0xC6, 0x40);    /* data slicer                                          */ 
	ret |= vpx32xx_writeI2C (client, 0xC5, 0x07);    /* filter coefficient                                   */
	ret |= vpx32xx_writeI2C (client, 0xC9, 0x01);    /* standard ... pal, ttx                                */
	ret |= vpx32xx_writeI2C (client, 0xBB, 0x27);    /* clock run-in and framing code reference high         */
	ret |= vpx32xx_writeI2C (client, 0xBC, 0x55);    /* clock run-in and framing code reference mid          */
	ret |= vpx32xx_writeI2C (client, 0xBD, 0x55);    /* clock run-in and framing code reference low          */
	ret |= vpx32xx_writeI2C (client, 0xC1, 0xBE);    /* ttx bitslicer frequency LSB	... wst pal             */
	ret |= vpx32xx_writeI2C (client, 0xC2, 0x0A);    /* ttx bitslicer frequency MSB                          */
	ret |= vpx32xx_writeI2C (client, 0xCF, 0xAB);    /* output mode	vid.lin data output, 64-byte mode disable */

	ret |= vpx32xx_writeFP (client, 0x0134, 0x013F); /* Start line even field  319  */
	ret |= vpx32xx_writeFP (client, 0x0135, 0x014F); /* End line even field    335  */
	ret |= vpx32xx_writeFP (client, 0x0136, 0x0006); /* Start line odd field     6  */
	ret |= vpx32xx_writeFP (client, 0x0137, 0x0016); /* End line odd field      22  */

	ret |= vpx32xx_writeFP (client, 0x0139, 0x0000); /* Slicer data size 0 (corresp. to 64)     */
	ret |= vpx32xx_writeFP (client, 0x0140, 0x0C60); /* Register for Control and Latching       */ 
	ret |= vpx32xx_writeFP (client, 0x0150, 0x0140); /* Format selection 	disable splitting  */
	ret |= vpx32xx_writeFP (client, 0x0154, 0x0000); /* Multiplexer/Multipurpose output         */ 

	ret |= vpx32xx_writeI2C (client, 0xAA, 0x10);    /* Low power mode, bit 4 connected to TDO-pin  */
	
	ret |= vpx32xx_writeFP (client, 0x0138, 0x0803); /* Control VBI-Window, sliced data, vbi enable */ 
	ret |= vpx32xx_writeFP (client, 0x0030, 0x0666); /* Color Processing                            */
	return ret; 
}

/* VBI raw data on VPX3225 */
static int vpx3225_init_vbi_raw (struct i2c_client *client)
{
	int ret = 0;

	/* vbi-raw initialisation */

	/* Load Table for Window #1 and #2 */
	ret |= vpx32xx_writeFP (client, 0x0121, 0x0C00); /* Vertical Lines In (window #1 disabled) */
	ret |= vpx32xx_writeFP (client, 0x012B, 0x0C00); /* Vertical Lines In (window #2 disabled) */

	/* Load Table for VBI-Window */
	ret |= vpx32xx_writeFP (client, 0x0134, 0x013F);	/* Start line even field 319            */
	ret |= vpx32xx_writeFP (client, 0x0135, 0x014F);	/* End line even field   335            */
	ret |= vpx32xx_writeFP (client, 0x0136, 0x0006);	/* Start line odd field    6            */
	ret |= vpx32xx_writeFP (client, 0x0137, 0x0016);	/* End line odd field     22            */
	ret |= vpx32xx_writeFP (client, 0x0139, 0x0000);	/* Slicer data size 0 (corresp. to 64)  */

	/* Control Word */
	ret |= vpx32xx_writeFP (client, 0x0140, 0x0C60);	/* Register for Control and Latching    */ 

	/* Formatter */
	ret |= vpx32xx_writeFP (client, 0x0150, 0x0004);	/* Format selection                     */

	/* Length/Polarity of HREF, VREF, FIELD   */	
	ret |= vpx32xx_writeFP (client, 0x0153, 0x0027); 

	/* Output Multiplexer */
	ret |= vpx32xx_writeFP (client, 0x0154, 0x0000);	/* Multiplexer/Multipurpose output      */ 

	/* Output */
	ret |= vpx32xx_writeI2C (client, 0xAA, 0x10);    /* Low power mode, bit 4 connected to TDO-pin */
	ret |= vpx32xx_writeI2C (client, 0xF2, 0x0F);    /* Output enable (A,B,pixclk,href,vref,field,vact,llc,llc2 */

	ret |= vpx32xx_writeFP (client, 0x0138, 0x0801); /* Control VBI-Window, raw data, vbi enable */ 

	/* Color Processing */	
	ret |= vpx32xx_writeFP (client, 0x0030, 0x0666); /* disable acc */

	return ret; 
}
#endif

/*****************************************************************************/

static void vpx32xx_inc_use (struct i2c_client *client)
{
	DPRINTK ("vpx32xx_inc_use()\n");
	MOD_INC_USE_COUNT;
}

static void vpx32xx_dec_use (struct i2c_client *client)
{
	DPRINTK ("vpx32xx_dec_use()\n");
	MOD_DEC_USE_COUNT;
}

/*****************************************************************************/

static int vpx32xx_attach (struct i2c_adapter *adap, int addr, unsigned short flags, int kind)
{
	struct i2c_client *client;
	struct vpx32xx *decoder;
	int ret = 0;

	if (NULL == (client = kmalloc (sizeof (struct i2c_client) + sizeof (struct vpx32xx), GFP_KERNEL))) 
		return -ENOMEM;
	memset (client, 0, sizeof (struct i2c_client) + sizeof (struct vpx32xx));

	decoder = (struct vpx32xx *) client->data = &(client[1]);

	strcpy (client->name, "vpx32xx");
	client->id = id++;
	client->flags = 0;
	client->addr = addr;
	client->adapter = adap;
	client->driver = &driver;

	if (0 == (ret = i2c_attach_client (client))) {
		printk (KERN_INFO "vpx32xx.o: vpx32xx attached\n");
	} else {
		printk (KERN_WARNING "vpx32xx.o: vpx32xx attach FAILED!\n");
		kfree (client);
		return ret;
	}

	/* chip type detection */
	if (0xEC == vpx32xx_readI2C (client, 0x00)) {
		decoder->part_number = (vpx32xx_readI2C (client, 0x02) << 8) + vpx32xx_readI2C (client, 0x01);

		printk ("vpx32xx.o: vpx32xx chip detection: MICRONAS INTERMETALL ");

		switch (decoder->part_number) {
			case VPX3225D: printk ("VPX3225D "); break;
			case VPX3224D: printk ("VPX3224D "); break;
			case VPX3220A: printk ("VPX3220A "); break;
			case VPX3216B: printk ("VPX3216B "); break;
			case VPX3214C: printk ("VPX3214C "); break;
			default:
				printk ("chip detected but unsupported by this driver\n");
				i2c_detach_client (client);
				kfree(client);
				return -1;
		}
		printk ("chip detected and attached \n");
	} else {
		printk ("vpx32xx.o: vpx32xx chip detection: no MICRONAS INTERMETALL chip found!\n");
		i2c_detach_client (client);
		kfree (client);
		return -1;
	}

	/* load most common values into the parameter-list */
	decoder->norm = VIDEO_MODE_PAL;
	decoder->brightness1 = decoder->brightness2 = 32768;
	decoder->contrast1   = decoder->contrast2   = 32768;
	decoder->saturation  = 32768;
	decoder->hue         = 32768;
	decoder->input       = 2;

	switch (decoder->part_number) {	/* set fp-ram register addresses */
		case VPX3225D: case VPX3224D: 
			decoder->fpsta = 0x35; decoder->fprd  = 0x36;
			decoder->fpwr  = 0x37; decoder->fpdat = 0x38;
			ret |= vpx322x_set_video_norm (client, decoder->norm);
			ret |= vpx32xx_selmux (client, decoder->input);
			ret |= vpx322x_set_picture (client);
			break;
		case VPX3220A: case VPX3216B: case VPX3214C:
			decoder->fprd  = 0x26; decoder->fpwr  = 0x27;
			decoder->fpdat = 0x28; decoder->fpsta = 0x29;
			ret |= vpx32xx_set_video_norm (client, decoder->norm);
			ret |= vpx32xx_selmux (client, decoder->input);
			ret |= vpx32xx_set_picture (client);
			break;
		default:
			return -EINVAL;
	}

	if(0 != ret) {
		printk (KERN_WARNING "vpx32xx.o: vpx32xx attach FAILED!\n");
		i2c_detach_client (client);
		kfree (client);
		return ret;
	}

	MOD_INC_USE_COUNT;
	return 0;
}

static int vpx32xx_detach (struct i2c_client *client)
{
	int ret = 0;

	/* stop further output */
	vpx32xx_writeI2C (client, 0xF2, 0x00);

	if (0 == (ret = i2c_detach_client (client))) {
		printk (KERN_INFO "vpx32xx.o: vpx32xx detached\n");
	} else {
		printk (KERN_WARNING "vpx32xx.o: vpx32xx detach FAILED!\n");
	}
	kfree (client);

	MOD_DEC_USE_COUNT;
	return ret;
}

static int vpx32xx_command (struct i2c_client *client, unsigned int cmd, void *arg)
{
	struct vpx32xx *decoder = client->data;
	int ret = 0;

	switch (cmd) {
	case DECODER_GET_CAPABILITIES: {	/* get decoder capabilities */
		struct video_decoder_capability *dc = arg;
		DPRINTK (KERN_INFO "vpx32xx.o: vpx32xx_command: DECODER_GET_CAPABILITIES\n");
		dc->flags = VIDEO_DECODER_PAL   |
			    VIDEO_DECODER_NTSC  |
			    VIDEO_DECODER_SECAM;
		dc->inputs  = VPX32XX_MAX_INPUT;
		dc->outputs = VPX32XX_MAX_OUTPUT;
		return ret;
	}
				
	case DECODER_GET_STATUS: {	/* get decoder status */
		int res = 0, status = 0;
		DPRINTK (KERN_INFO "vpx32xx.o: vpx32xx_command: DECODER_GET_STATUS ");
		if (VPX3225D == decoder->part_number || 
		    VPX3224D == decoder->part_number   ) {
			status = vpx32xx_readFP (client, 0x13);
		}
		DPRINTK ("(vpx32xx-status=0x%03x)\n", status);

		if (0 > status)
			return status; 
		if ((status & 4) == 0)
			res |= DECODER_STATUS_GOOD;
		res |= DECODER_STATUS_COLOR;

		switch (decoder->norm) {
		case VIDEO_MODE_PAL:
			res |= DECODER_STATUS_PAL;
			break;
		case VIDEO_MODE_NTSC:
			res |= DECODER_STATUS_NTSC;
			break;
		case VIDEO_MODE_SECAM:
			res |= DECODER_STATUS_SECAM;
			break;
		default:
		}
		*(int *) arg = res;
		return ret;
	}
	
	case DECODER_SET_NORM: {	/* set video norm */
		__u16 video_norm = *(__u16 *) arg;
		DPRINTK (KERN_INFO "vpx32xx.o: vpx32xx_command: DECODER_SET_NORM (norm=%d)\n", video_norm);
		switch (decoder->part_number) {
		case VPX3225D: case VPX3224D:
			return vpx322x_set_video_norm (client, video_norm);
		case VPX3220A: case VPX3216B: case VPX3214C:
			return vpx32xx_set_video_norm (client, video_norm);
		default:
			return -EINVAL;
		}
	}
		
	case DECODER_SET_INPUT: {	/* set video channel input */
		int video_input = *(int*)arg;
		DPRINTK (KERN_INFO "vpx32xx.o: vpx32xx_command: DECODER_SET_INPUT (video channel input=%d)\n", video_input);
		if ((VPX32XX_MAX_INPUT <= video_input) || (video_input < 0)) /* very important test !! */
			return -EINVAL;
		return vpx32xx_selmux (client, video_input);
	}

	case DECODER_SET_OUTPUT: {	/* set decoder output - not supported here */
		DPRINTK (KERN_INFO "vpx32xx.o: vpx32xx_command: DECODER_SET_OUTPUT not supported\n");
		return -EOPNOTSUPP;
	}
		
	case DECODER_ENABLE_OUTPUT: {	/* decoder enable output */
		int enable = * (int *) arg;
		DPRINTK (KERN_INFO "vpx32xx.o: vpx32xx_command: DECODER_ENABLE_OUTPUT\n");
		if (enable) {
			vpx32xx_writeI2C (client, 0xAA, 0x00);  /* Power mode                      */
			vpx32xx_writeFP (client, 0x22, 0x0E);   /* Start of active video           */
			vpx32xx_writeI2C (client, 0xF2, 0x0D);  /* Output Enable Video Port A only */
			vpx32xx_writeI2C (client, 0xF8, 0x24);  /* Driver strength Typ A           */
			vpx32xx_writeI2C (client, 0xF9, 0x04);  /* Driver strength Typ B           */
			vpx32xx_writeFP (client, 0x153, 0x00);  /* HREF and VREF control           */
			vpx32xx_writeFP (client, 0x150, 0x601); /* enable ITU-R656                 */
			vpx32xx_writeFP (client, 0x154, 0x200); /* Output Multiplexer              */
			vpx32xx_writeFP (client, 0x151, 0x00);  /* Start position                  */
			vpx32xx_writeFP (client, 0x152, 800);   /* End position                    */
			vpx32xx_writeFP (client, 0x130, 0x204); /* Peaking/Coring                  */
			vpx32xx_writeFP (client, 0x30, 0xBE0);  /* ACC reference level             */
			vpx32xx_writeFP (client, 0x140, 0x800); /* Control/Latching                */
		} else {
			vpx32xx_writeI2C (client, 0xF2, 0x00);
		}
		return 0;
	}
		
	case DECODER_SET_PICTURE: {	/* set brightness, contrast, saturation, and hue = 0, ..., 65535 */
		struct video_picture *pic = arg;
		DPRINTK (KERN_INFO "vpx32xx.o: vpx32xx_command: DECODER_SET_PICTURE (brightness=%d, hue=%d, color=%d, contrast=%d)\n", pic->brightness, pic->hue, pic->colour, pic->contrast);
		decoder->brightness1 = pic->brightness;
		decoder->brightness2 = pic->brightness;
		decoder->contrast1   = pic->contrast;
		decoder->contrast2   = pic->contrast;
		decoder->saturation  = pic->colour;
		decoder->hue         = pic->hue;

		switch (decoder->part_number) {
		case VPX3225D: case VPX3224D:
			return vpx322x_set_picture (client);
		case VPX3220A: case VPX3216B: case VPX3214C:
			return vpx32xx_set_picture (client);
		default:
			return -EINVAL;
		}
	}

#if 0
	case DECODER_SET_VBI: {		/* set parameters for vbi mode */
		DPRINTK (KERN_INFO "vpx32xx.o: vpx32xx_command: DECODER_SET_VBI\n");
		if (VPX3225D == decoder->part_number)	/* vbi_sliced only for VPX3225 */
			return vpx3225_init_vbi_sliced (client);
		return vpx3225_init_vbi_raw (client);
		return -EINVAL;
	}
#endif

	case DECODER_DUMP: {
		DPRINTK (KERN_INFO "vpx32xx.o: vpx32xx_command: DECODER_DUMP\n");
		if ((VPX3225D == decoder->part_number) || (VPX3224D == decoder->part_number))
			return vpx322x_dump (client);
		return -EINVAL;
	}

	default:
		DPRINTK (KERN_INFO "vpx32xx.o: vpx32xx_command: unknown ioctl (cmd=0x%x)\n", cmd);
		return -ENOIOCTLCMD;
	}
}

static int vpx32xx_probe (struct i2c_adapter *adap)
{
	return i2c_probe (adap, &addr_data, vpx32xx_attach); 
}

static struct i2c_driver driver = {
	name:            "vpx32xx",
	id:              I2C_DRIVERID_VPX32XX,
	flags:           I2C_DF_NOTIFY,
	attach_adapter:  vpx32xx_probe,
	detach_client:   vpx32xx_detach,
	command:         vpx32xx_command,
	inc_use:	 vpx32xx_inc_use,
	dec_use:	 vpx32xx_dec_use
};

/*****************************************************************************/

MODULE_DESCRIPTION ("vpx32xx - i2c-driver for MICRONAS INTERMETALL VPX32XX video pixel decoder");
MODULE_AUTHOR      ("Peter Schlaf");
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 4, 10)
MODULE_LICENSE     ("GPL");
#endif

int vpx32xx_init_module (void)
{
	i2c_add_driver (&driver);
	return 0;
}

void vpx32xx_cleanup_module (void)
{
	i2c_del_driver (&driver);
}

module_init (vpx32xx_init_module);
module_exit (vpx32xx_cleanup_module);
